#include "drawable-node.h"
#include <string.h>
#include "wsint.h"

#include <X11/Xlib.h>
#include <X11/extensions/Xfixes.h>

#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#   define PIXEL_TYPE GL_UNSIGNED_BYTE
#else
#   define PIXEL_TYPE GL_UNSIGNED_INT_8_8_8_8_REV
#endif

static void drawable_node_update_pixmap (DrawableNode *node);

static guint
power_of_two_greater_than (guint in)
{
    guint result = 1;
    
    while (result < in)
	result *= 2;
    
    return result;
}

static guint
power_of_two_less_than_or_equal (guint in)
{
    guint result;

    g_return_if_fail (in > 0);
    
    result = power_of_two_greater_than (in);

    if (result > in)
	return result / 2;
    else
	return result;
}

static WsShmPixmap *
get_shm_pixmap (WsDrawable *drawable,
		int	    width,
		int	    height,
		WsFormat    format)
{
    static WsShmPixmap *cache;
    static int cached_width;
    static int cached_height;
    static WsFormat cached_format;
    
    if (!cache || cached_width < width ||
	cached_height < height || cached_format != format)
    {
	if (cache)
	{
	    ws_shm_pixmap_unref (cache);
	    cache = NULL;
	}
	
	cache = ws_shm_pixmap_new (drawable, width, height, format);

	/* FIXME: we should do this properly at some point */
	g_assert (cache);
	
	cached_width = width;
	cached_height = height;
	cached_format = format;
    }
    
    return cache;
}

static gboolean
has_child_prefixed (WsWindow *window,
		    char *child_title)
{
    GList *list;
    GList *children = ws_window_query_subwindows (window);
    int child_len = strlen (child_title);
    
    for (list = children; list ; list = list->next)
    {
	char *title = ws_window_query_title (list->data);
	
	if (title && strncmp (child_title, title, child_len) == 0)
	    return TRUE;
    }
    
    return FALSE;
}

static void
fill_junk (guint32 *mem, gsize size)
{
    int i;
    
    for (i = 0; i < size/4; ++i)
	mem[i] = g_random_int ();
}

static int counter;
WsWindow *debug_window;

static guchar *
create_buffer (WsShmPixmap *pixmap, WsFormat format,
	       guint width, guint height)
{
    guchar *buffer;
    const guchar *src;
    int i, j;
    guint bytes_per_row;
    guint bytes_per_pixel;
    guint pot_width, pot_height;

    bytes_per_row = ws_shm_pixmap_get_bytes_per_row (pixmap);
    src = ws_shm_pixmap_get_data (pixmap);

    /* FIXME */
    bytes_per_pixel = 4; 

    pot_width = power_of_two_greater_than (width);
    pot_height = power_of_two_greater_than (height);
    
    buffer = g_malloc0 (4 * pot_height * pot_width);
#if 0
    fill_junk ((guint32 *)buffer, 4 * pot_height * pot_width);
#endif
    
    for (j = 0; j < height; ++j)
    {
	const guchar *src_row = src + (j % height) * bytes_per_row;
	guint32 *dest_row = (guint32 *)(buffer + j * 4 * pot_width);
	
	for (i = 0; i < width; ++i)
	{
	    const guchar *pixel = src_row + (i % width) * bytes_per_pixel;

	    dest_row[i] = *(guint32 *)pixel;

	    if (format == WS_FORMAT_RGB_32 ||
		format == WS_FORMAT_RGB_24)
	    {
		dest_row[i] |= 0x77ffffff;
	    }
	    else
	    {
		dest_row[i] = 0x77000077;
	    }

	    {
		/* FIXME: unpremultiplying - need to figure out how to do that
		 * efficiently
		 */
		guchar *dst = (guchar *)(dest_row + i);
		
		if (dst[3] != 0)
		{
		    double alpha = dst[3] / 255.0;

		    dst[0] = dst[0] / alpha;
		    dst[1] = dst[1] / alpha;
		    dst[2] = dst[2] / alpha;
		}
	    }
	}
    }
    
    return buffer;
}

static void
print_region (WsRegion *region, const char *name)
{
    XRectangle *rects;
    int i, n_rects;
    
    rects = XFixesFetchRegion (region->ws->xdisplay, region->xregion, &n_rects);
    g_print ("region \"%s\":\n", name);
    for (i = 0; i < n_rects; ++i)
	g_print ("  %d %d %d %d\n", rects[i].x, rects[i].y, rects[i].width, rects[i].height);
    XFree (rects);
}

typedef struct Tile Tile;
struct Tile
{
    gint x;
    gint y;
    gint width;
    gint height;

    GLuint texture;
};

static guchar *
convert_data (WsBits *bits,
	      WsFormat format,
	      guint   x,
	      guint   y,
	      guint   width,
	      guint   height)
{
    int i, j;
    guint bytes_per_line;
    guint bytes_per_pixel;
    guchar *result;
    const guchar *src;

    src = ws_bits_get_info (bits, NULL, NULL,
			    &bytes_per_pixel,
			    &bytes_per_line);

    result = g_malloc0 (4 * width * height);

    for (j = 0; j < height; ++j)
    {
	const guchar *src_row = src + x * bytes_per_pixel + (y + j) * bytes_per_line;
	guint32 *dest_row = (guint32 *)(result + j * 4 * width);
	
	for (i = 0; i < width; ++i)
	{
	    const guchar *pixel = src_row + i * bytes_per_pixel;

	    dest_row[i] = *(guint32 *)pixel;

	    if (format == WS_FORMAT_RGB_32 ||
		format == WS_FORMAT_RGB_24)
	    {
		dest_row[i] &= 0x00ffffff;
		dest_row[i] |= 0xff000000;
	    }

#if 0
	    {
		/* FIXME: unpremultiplying - need to figure out how to do that
		 * efficiently
		 */
		guchar *dst = (guchar *)(dest_row + i);
		
		if (dst[3] != 0)
		{
		    double alpha = dst[3] / 255.0;

		    dst[0] = dst[0] / alpha;
		    dst[1] = dst[1] / alpha;
		    dst[2] = dst[2] / alpha;
		}
	    }
#endif
	}
    }
    
    return result;
}

static GList *
create_tiles (WsRectangle *rectangle)
{
    /* FIXME: at some point we should make
     * sure to create overlapping tiles
     */
    GList *columns = NULL;
    GList *tiles = NULL;
    GList *list;

    guint x = rectangle->x;
    guint width = rectangle->width;

    /* First split the rectangle horizontally into columns */
    while (width > 0)
    {
	Tile *column = g_new0 (Tile, 1);
	column->x = x;
	column->y = rectangle->y;
	column->width = power_of_two_less_than_or_equal (width);
	column->height = rectangle->height;

	x += column->width;
	width -= column->width;

	columns = g_list_prepend (columns, column);
    }

    /* Then split each column vertically */
    for (list = columns; list != NULL; list = list->next)
    {
	Tile *column = list->data;
	
	guint y = column->y;
	guint height = column->height;

	while (height > 0)
	{
	    Tile *tile = g_new0 (Tile, 1);
	    tile->x = column->x;
	    tile->y = y;
	    tile->width = column->width;
	    tile->height = power_of_two_less_than_or_equal (height);

	    y += tile->height;
	    height -= tile->height;

	    tiles = g_list_prepend (tiles, tile);
	}

	g_free (column);
    }

    return tiles;
}

static void
dump_error (const char *where)
{
    GLenum err = glGetError ();

    if (err != GL_NO_ERROR)
	g_print ("error: %s   %s\n", where, gluErrorString (err));
}

static void
update_tile (Tile *tile,
	     WsFormat format,
	     WsBits *bits,
	     WsRectangle *area,
	     gboolean first)
{
    guint32 *buffer;
    
    glBindTexture (GL_TEXTURE_2D, tile->texture);
    
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP);

    if (first)
    {
	buffer = (guint32 *)convert_data (
	    bits, format,
	    0, 0, tile->width, tile->height);

	glTexImage2D (GL_TEXTURE_2D, 0, GL_RGBA,
		      tile->width, tile->height, 0 /* border */,
		      GL_BGRA, PIXEL_TYPE, buffer);
    }
    else
    {
	buffer = (guint32 *)convert_data (
	    bits, format, 0, 0, area->width, area->height);

	glTexSubImage2D (GL_TEXTURE_2D, 0,
			 area->x - tile->x,
			 area->y - tile->y,
			 area->width, area->height,
			 GL_BGRA, PIXEL_TYPE,
			 buffer);
    }
    
    g_free (buffer);
}

static void
print_rect (const char *caption,
	    WsRectangle *rect)
{
    g_print ("%s:   %d %d %d %d\n", caption, rect->x, rect->y, rect->width, rect->height);
}

static void
create_texture2 (DrawableNode *node,
		 gboolean      get_all,
		 WsRectangle  *geometry)
{
    XImage *image;
    const guint32 *data;
    guint width;
    guint height;
    guint bytes_per_pixel;
    guint bytes_per_line;
    GList *list;
    WsFormat format;
    WsRectangle *free_me = NULL;
    WsRectangle *rects;
    guint n_rects;
    int i;

    ws_begin_error_trap (node->drawable->ws);
 
    if (!node->tiles)
	node->tiles = create_tiles (geometry);

    /* We need a sync here so that we will know that the data has been
     * copied. FIXME: could most likely be done more efficiently.
     *
     * We may even need a getimage type thing to make sure the stuff
     * has been flushed to the card.
     */
    ws_sync (ws_drawable_get_ws ((WsDrawable *)node->pixmap));
    
    glXWaitX();
    
    format = ws_window_query_format ((WsWindow *)node->drawable);
    
    ws_end_error_trap (node->drawable->ws);

    if (get_all)
    {
	rects = geometry;
	n_rects = 1;
    }
    else
    {
	free_me = rects =
	    ws_region_query_rectangles (node->damaged_region, &n_rects);
    }

    for (i = 0; i < n_rects; ++i)
    {
	WsRectangle *damage_rect = &(rects[i]);

	for (list = node->tiles; list != NULL; list = list->next)
	{
	    Tile *tile = list->data;
	    guint32 *buffer;
	    WsRectangle tile_rect;
	    WsRectangle intersection;
	    gboolean first = FALSE;

	    if (!tile->texture)
	    {
		glGenTextures (1, &tile->texture);
		first = TRUE;
	    }

	    tile_rect.x = tile->x;
	    tile_rect.y = tile->y;
	    tile_rect.width = tile->width;
	    tile_rect.height = tile->height;
	    
	    if (ws_rectangle_intersect (
		    damage_rect, &tile_rect, &intersection))
	    {
		WsBits *bits = ws_drawable_get_bits ((WsDrawable *)node->pixmap, &intersection);
		if (bits)
		    update_tile (tile, format, bits, &intersection, first);
		ws_bits_free (bits);
	    }
	}
    }

    if (free_me)
	g_free (free_me);
    
    node->we_have_a_texture = 0xc001;

    g_assert (node->tiles);
    
    debug_window = (WsWindow *)node->drawable;
}

static void
delete_tiles (DrawableNode *node)
{
    GList *list;

    for (list = node->tiles; list != NULL; list = list->next)
    {
	Tile *tile = list->data;

	glDeleteTextures (1, &tile->texture);

	g_free (tile);
    }

    g_list_free (node->tiles);
    node->tiles = NULL;

    node->we_have_a_texture = 0xBAD;
}

static void
drawable_repair (DrawableNode *node)
{
    WsDrawable *drawable = node->drawable;
    gboolean get_all;
    
    /* FIXME: maybe at some point we should get
     * smart and do incremental update.
     */

    if (node->need_new_pixmap)
    {
	ws_begin_error_trap (drawable->ws);
	
	if (node->pixmap)
	    ws_pixmap_free (node->pixmap);

	node->pixmap = ws_window_name_pixmap ((WsWindow *)drawable);
	
	ws_end_error_trap (drawable->ws);
    }
    else
    {
	if (!node->damaged_region)
	    return;
    }
    
    ws_begin_error_trap (drawable->ws);
    
    if (node->pixmap)
    {
	WsRectangle geometry;
	
	if (ws_drawable_query_geometry ((WsDrawable *)node->pixmap,
					&geometry))
	{
	    create_texture2 (node, node->need_new_pixmap, &geometry);
	    node->need_new_pixmap = FALSE;
	}
	else
	{
	    node->pixmap = NULL;
	}
    }
    
    if (node->damaged_region)
	ws_region_unref (node->damaged_region);
    
    ws_end_error_trap (drawable->ws);

    node->damaged_region = NULL;
    
    node_queue_repaint ((Node *)node);
}

static void
on_damage (WsDrawable *drawable,
	   gpointer    data)
{
    DrawableNode *node = data;
    WsRegion *region;

    ws_begin_error_trap (drawable->ws);
    
    region = ws_drawable_undamage (drawable);

    if (node->drawable->xid == 0xe0002e)
    {
	g_print ("nautilus damage\n");
	print_region (region, "nautilus");
    }
    
#if 0
    if (has_child_prefixed ((WsWindow *)drawable, "X-Chat"))
    {
	g_print ("damaged region is\n");

    }
#endif
    
    ws_end_error_trap (drawable->ws);

    if (node->damaged_region)
    {
	ws_region_union (node->damaged_region, region);
	ws_region_unref (region);
    }
    else
    {
	node->damaged_region = region;
    }
}

static void
deform (DrawableNode *node,
	double time,
	double inx,
	double iny,
	double *outx,
	double *outy)
{
    if (node->deformation_func)
    {
	DrawableDeformationFunc deformation_func = node->deformation_func;
	gpointer data = node->deformation_data;
	
	deformation_func (time, inx, iny, outx, outy, data);
    }
    else
    {
	*outx = inx;
	*outy = iny;
    }
}

static void
render_tile (DrawableNode *node,
	     double	   time,
	     int	   x,
	     int	   y,
	     WsScreen     *screen,
	     Tile         *tile)
{
    double scr_w, scr_h;
    double tw, th;
    double vw, vh;
    double vx, vy;
    
    scr_w = ws_screen_get_width (screen);
    scr_h = ws_screen_get_height (screen);

    tw = 1.0;
    th = 1.0;

    vw = tile->width / scr_w;
    vh = tile->height / scr_h;

    vx = (x + tile->x) / scr_w;
    vy = (y + tile->y) / scr_h;

#define GRID_SIZE 16
    {
	int n_grid_cols_x;
	int n_grid_cols_y;
	int i, j;

#if 0
	g_print ("drawing %d\n", tile->texture);
#endif
	glBindTexture (GL_TEXTURE_2D, tile->texture);
	
	glEnable (GL_TEXTURE_2D);
	glEnable (GL_BLEND);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f (1.0, 1.0, 1.0, 1.0);

	n_grid_cols_x = MAX (tile->width / GRID_SIZE, 1);
	n_grid_cols_y = MAX (tile->height / GRID_SIZE, 1);
	
	for (i = 0; i < n_grid_cols_x; ++i)
	{
	    for (j = 0; j < n_grid_cols_y; ++j)
	    {
		double tx0, ty0, tx1, ty1;
		double wx0, wy0, wx1, wy1;

		/* The transformed points */
		double p0x, p0y;
		double p1x, p1y;
		double p2x, p2y;
		double p3x, p3y;
		    
		tx0 = (tw / n_grid_cols_x) * i;
		tx1 = (tw / n_grid_cols_x) * (i + 1);
		ty0 = (th / n_grid_cols_y) * j;
		ty1 = (th / n_grid_cols_y) * (j + 1);

		wx0 = vx + (vw / n_grid_cols_x) * i;
		wx1 = vx + (vw / n_grid_cols_x) * (i + 1);
		wy0 = 1.0 - (vy + (vh / n_grid_cols_y) * j);
		wy1 = 1.0 - (vy + (vh / n_grid_cols_y) * (j + 1));

		deform (node, time, wx0, wy0, &p0x, &p0y);
		deform (node, time, wx1, wy0, &p1x, &p1y);
		deform (node, time, wx1, wy1, &p2x, &p2y);
		deform (node, time, wx0, wy1, &p3x, &p3y);

		glBegin (GL_QUADS);
		glTexCoord2f (tx0, ty0); glVertex3f (p0x, p0y, 0.0);
		glTexCoord2f (tx1, ty0); glVertex3f (p1x, p1y, 0.0);
		glTexCoord2f (tx1, ty1); glVertex3f (p2x, p2y, 0.0);
		glTexCoord2f (tx0, ty1); glVertex3f (p3x, p3y, 0.0);
		glEnd ();
	    }
	}
    }
}

static void
drawable_node_render (Node *node)
{
    DrawableNode *drawable_node = (DrawableNode *)node;
    WsRectangle rect;
    WsScreen *screen;
    double scr_w, scr_h;
    double tw, th;
    double vx, vy;
    double vw, vh;
    gboolean compositable = TRUE;
    GList *list;

    ws_begin_error_trap (drawable_node->drawable->ws);
    
    if (!ws_window_query_viewable ((WsWindow *)(drawable_node->drawable)) )
    {
	if (drawable_node->need_new_pixmap)
	    compositable = FALSE;
    }

    if (ws_window_query_input_only ((WsWindow *)drawable_node->drawable))
	compositable = FALSE;

    ws_end_error_trap (drawable_node->drawable->ws);

    if (!compositable || !drawable_node->viewable)
	return;

    drawable_repair (drawable_node);
    
    ws_begin_error_trap (drawable_node->drawable->ws);
    
    ws_drawable_query_geometry (drawable_node->drawable, &rect);
    screen = ws_drawable_query_screen (drawable_node->drawable);

    ws_end_error_trap (drawable_node->drawable->ws);

    if (!screen)
	return;

    gdouble time = g_timer_elapsed (drawable_node->timer, NULL);

    for (list = drawable_node->tiles; list != NULL; list = list->next)
    {
	Tile *tile = list->data;

	render_tile (drawable_node, time, rect.x, rect.y, screen, list->data);
    }
}

static void
drawable_node_compute_extents (Node    *node,
			       Extents *extents)
{
    extents->x = 0.0;
    extents->y = 0.0;
    extents->z = 0.0;
    extents->width = 1.0;
    extents->height = 1.0;
    extents->depth = 1.0;
}

static void
on_configure (WsWindow *drawable,
	      int x,
	      int y,
	      int width,
	      int height,
	      int bw,
	      WsWindow *above,
	      gboolean override_redirect,
	      gpointer data)
{
    DrawableNode *node = data;
#if 0
    g_print ("configure\n");
    
    g_print ("%d %d %d %d\n", geometry.x, geometry.y,
	     geometry.width, geometry.height);
#endif
    
    delete_tiles (node);
    node->need_new_pixmap = TRUE;
}

DrawableNode *
drawable_node_new (WsDrawable *drawable)
{
    DrawableNode *node;
    WsRectangle rectangle;

    g_return_val_if_fail (drawable != NULL, NULL);

    node = g_new0 (DrawableNode, 1);

#if 0
    g_print ("drawable node new %p\n", node);
#endif

    node->node.render = drawable_node_render;
    node->node.compute_extents = drawable_node_compute_extents;

    node->drawable = drawable;
    node->tiles = NULL;
    node->we_have_a_texture = 0xBAD;

    ws_init_composite (ws_drawable_get_ws (drawable));
    
    /* FIXME: we are assuming it here again */
    ws_begin_error_trap (drawable->ws);

    node->need_new_pixmap = TRUE;
    node->pixmap = NULL;

    node->timer = g_timer_new ();
    
    ws_end_error_trap (drawable->ws);
    
    node->damaged_region = NULL;
    
    ws_init_damage (ws_drawable_get_ws (drawable));
    
    ws_drawable_set_damage_callback (drawable, on_damage, node);

    ws_window_set_configure_callback ((WsWindow *)drawable, on_configure, node);

    node->viewable = TRUE;
    
    return node;
}

void
drawable_node_set_viewable (DrawableNode *node,
			    gboolean	  viewable)
{
    node->viewable = viewable;

    node_queue_repaint ((Node *)node);
}

void
drawable_node_unref (DrawableNode *node)
{
    ws_drawable_free (node->drawable);

    if (node->tiles)
    {
	delete_tiles (node);
    }

    if (node->damaged_region)
    {
	ws_region_unref (node->damaged_region);
	node->damaged_region = NULL;
    }

#if 0
    node_unref ((Node *)node);
#endif
}

void
drawable_node_set_deformation_func (DrawableNode *		node,
				    DrawableDeformationFunc	func,
				    gpointer			data)
{
    node->deformation_func = func;
    node->deformation_data = data;
}
						 
