#ifndef _RENDER_H
#define _RENDER_H

#include <glib.h>

typedef struct Node Node;

/* A RenderNode will assume that it has a coordinate system
 * with properties such that
 *
 *	- (0, 0, 0) is in the bottom left corner of the screen
 *	- (0, 1, 0) is in the top left corner of the screen
 *      - (1, 0, 0) has the same distance to (0, 0, 0) as (0, 1, 0)
 *      - (0, 0, 1) has (visually) the same distance to (0, 0, 0)
 *
 *
 * A RenderNode has a 'render' method that draws the thing.
 *
 * Possibly later: A RenderNode has an 'extents' method that returns a
 * bounding box.
 */

/* Example of possibly useful RenderNodes:
 * 
 * 
 * Cube: Has six children that are rendered on the six faces of the cube
 *	Cube can be spun around axes.
 *		cube_rotate_{x,y,z} ()
 *
 * Sphere: Renders its child(s 2D projection) on the surface of a sphere:
 *
 *	implementation:
 *		compute extents of child
 *		split extent into number of small boxes
 *		for each box
 *			push
 *			set scissor/clip to box
 *			set projection so that drawing ends
 *				up on a square on the cube
 *			(* render)
 *			pop()
 *
 * Canvas: Renders a number of stacked 2D children.
 *
 * Distortion: Splits child into a number of rectangles,
 *
 * Window: renders a window.
 *
 * Other useful ones: Translator, Rotator, Box (puts more children aside each other).
 *
 */

/*
 *	Silly animation:
 *		Grid:	  Renders its (n * m) children in an (n x m) grid
 *		Clip:     Takes one child and clips out a given retangle (or box).
 *					(Note: use ClipPlane)
 *
 *              Add cubes to grid, each cube has a Clip of the desktop.
 *
 */ 

/* We'll need the concept of 'extent'. A node must report its extent -
 * otherwise, things like the cube won't know how much of its children
 * to render on its faces.
 *
 *      Extents {
 *
 *		double x,
 *		double y,
 *		double z,
 *		double width,
 *		double height,
 *		double depth
 *	};
 *
 * all the numbers can legally be 0, meaning that nothing is drawn.
 *
 * if only one dimension is 0, the extents is 2D. Does GL deal correctly
 * with clipping a 3D drawing to a 2D plane? We are going to need this
 * for a lot of things.
 *
 */

typedef struct Extents Extents;

struct Extents
{
    double x;
    double y;
    double z;
    double width;
    double height;
    double depth;
};

struct Node
{
    Extents extents;

    void (* render) (Node *node);
    void (* compute_extents) (Node    *node,
			      Extents *extents);
};

void node_get_extents (Node    *node,
		       Extents *extents);
void node_render (Node *node);
void node_queue_repaint (Node *node);
gdouble node_get_time (Node *node);

#endif /* _RENDER_H */
