/* Terminology
 *
 *     - create/destroy       - the X resource is created/destroyed
 *     - new/free/ref/unref   - the local proxy is created/destroyed
 *
 * Generally local proxies can exist even if the other resource doesn't
 * anymore. In that case they don't have an associated XID.
 *
 * And of course we are being async.
 *
 */

#include <X11/X.h>
#include <X11/Xutil.h>
#include <X11/extensions/Xcomposite.h>
#include <X11/extensions/Xdamage.h>
#include <X11/extensions/XTest.h>

#include <glib-object.h>

#include "ws.h"
#include "wsint.h"
#include "watch.h"

#if 0
typedef enum
{
    CORE,
    DAMAGE,
    SYNC
} Extension;

typedef struct
{
    guint	xtype;
    const char *name;
    int		drawable_offset;
    int		extenstion_event_offset;
} EventTranslation;

const EventTranslation translation_table[] =
{
    { KeyPress,   "KeyPress",   G_STRUCT_OFFSET (XKeyPressedEvent, window) } ,
    { KeyRelease, "KeyRelease", G_STRUCT_OFFSET (XKeyReleasedEvent, window) } ,
    { ButtonPress, "ButtonPress", G_STRUCT_OFFSET (XButtonEvent, window) } ,
    { ButtonRelease, "ButtonRelease", G_STRUCT_OFFSET (XButtonEvent, window) } ,
    { MotionNotify, "MotionNotify", G_STRUCT_OFFSET (XMotionEvent, window) } ,
    { EnterNotify, "EnterNotify", G_STRUCT_OFFSET (XEnterWindowEvent, window) } ,
    { LeaveNotify, "LeaveNotify", G_STRUCT_OFFSET (XLeaveWindowEvent, window) } ,
    { FocusIn, "FocusIn", G_STRUCT_OFFSET (XFocusInEvent, window) } ,
    { FocusOut, "FocusOut", G_STRUCT_OFFSET (XFocusOutEvent, window) } ,
    { KeymapNotify, "KeymapNotify", G_STRUCT_OFFSET (XKeymapEvent, window) } ,
    { Expose,           "Expose", G_STRUCT_OFFSET (XExposeEvent, window) } ,
    { GraphicsExpose,   "GraphicsExpose", G_STRUCT_OFFSET (XGraphicsExposeEvent, drawable) } ,
    { NoExpose,         "NoExpose", G_STRUCT_OFFSET (XNoExposeEvent, drawable) } ,
    { VisibilityNotify, "VisibilityNotify", G_STRUCT_OFFSET (XVisibilityEvent, window) } ,
    { CreateNotify, "CreateNotify", G_STRUCT_OFFSET (XCreateWindowEvent, window) } ,
    { DestroyNotify, "DestroyNotify", G_STRUCT_OFFSET (XDestroyWindowEvent, window) } ,
    { UnmapNotify, "UnmapNotify", G_STRUCT_OFFSET (XUnmapEvent, window) } ,
    { MapNotify, "MapNotify", G_STRUCT_OFFSET (XMapEvent, window) } ,
    { MapRequest, "MapRequest", G_STRUCT_OFFSET (XMapRequestEvent, window) } ,
    { ReparentNotify, "ReparentNotify", G_STRUCT_OFFSET (XReparentEvent, window) } ,
    { ConfigureNotify, "ConfigureNotify", G_STRUCT_OFFSET (XConfigureEvent, window) } ,
    { ConfigureRequest, "ConfigureRequest", G_STRUCT_OFFSET (XConfigureRequestEvent, window) } ,
    { GravityNotify, "GravityNotify", G_STRUCT_OFFSET (XGravityEvent, window) } ,
    { ResizeRequest, "ResizeRequest", G_STRUCT_OFFSET (XResizeRequestEvent, window) } ,
    { CirculateNotify, "CirculateNotify", G_STRUCT_OFFSET (XCirculateEvent, window) } ,
    { CirculateRequest, "CirculateRequest", G_STRUCT_OFFSET (XCirculateRequestEvent, window) } ,
    { PropertyNotify, "PropertyNotify", G_STRUCT_OFFSET (XPropertyEvent, window) } ,
    { SelectionClear, "SelectionClear", G_STRUCT_OFFSET (XSelectionClearEvent, window) } ,
    { SelectionRequest, "SelectionRequest", G_STRUCT_OFFSET (XSelectionRequestEvent, window) } ,
    { SelectionNotify, "SelectionNotify", G_STRUCT_OFFSET (XSelectionEvent, window) } ,
    { ColormapNotify, "ColormapNotify", G_STRUCT_OFFSET (XColormapEvent, window) } ,
    { ClientMessage, "ClientMessage", G_STRUCT_OFFSET (XClientMessageEvent, window) } ,
    { MappingNotify, "MappingNotify", G_STRUCT_OFFSET (XMappingNotifyEvent, window) },
};
#endif

static void
process_event (Ws      *ws,
	       XEvent  *xevent)
{
#if 0
    if (xevent->type == XMotionNotify)
	g_print ("motion\n");
    else
	g_print ("not motion\n");
#endif

    if (ws->damage.available &&
	xevent->type == ws->damage.event_base + XDamageNotify)
    {
	Window xwindow = ((XDamageNotifyEvent *)xevent)->drawable;
	WsDrawable *drawable = g_hash_table_lookup (
	    ws->xresources, (gpointer)xwindow);

	g_assert (drawable->damage_cb != (void *)0x01);

	g_assert (drawable->xid == xwindow);
	
	if (drawable)
	    _ws_drawable_process_event (drawable, xevent);
    }
    else if (xevent->type == ConfigureNotify)
    {
	Window xwindow = ((XConfigureEvent *)xevent)->window;
	WsWindow *window = g_hash_table_lookup (
	    ws->xresources, (gpointer)xwindow);

	if (window)
	    _ws_window_process_event (window, xevent);
    }
}

static void
process_events (gpointer data)
{
    Ws *ws = data;
    Display *xdisplay = ws->xdisplay;

    while (XPending (ws->xdisplay))
    {
	XEvent event;
	
	XNextEvent (xdisplay, &event);
	
	process_event (ws, &event);
    }
}

static void
init_ws (Ws *ws)
{
    int i;
    Display *xdisplay = ws->xdisplay;

    XSynchronize (xdisplay, True);
    
    ws->n_screens = ScreenCount (xdisplay);
    ws->screens = g_new0 (WsScreen *, ws->n_screens);

    for (i = 0; i < ws->n_screens; ++i)
	ws->screens[i] = ws_screen_get_from_number (ws, i);
    
    ws->xresources = g_hash_table_new (g_direct_hash, g_direct_equal);
    
    ws->composite.available = FALSE;
    ws->damage.available = FALSE;
    ws->shm.available = FALSE;
    ws->test.available = FALSE;
    ws->fixes.available = FALSE;
}

static ErrorHandlerFunc old_x_error_handler;
static GList *all_wses;

static void
pop_obsolete_handlers (Ws *ws,
		       int serial)
{
    Trap *new_top;

    new_top = ws->traps;
    while (new_top != NULL &&
	   new_top->end_serial != -1 &&
	   new_top->end_serial <= serial)
    {
	new_top = new_top->next;
    }

    while (ws->traps != new_top)
    {
	Trap *trap;

	trap = ws->traps;
	ws->traps = trap->next;

#if 0
	g_print ("popping: %d %d (%p)\n",
		 trap->begin_serial, trap->end_serial, ws->xdisplay);
#endif
	
	g_free (trap);
    }

    if (!ws->traps)
    {
	XSetErrorHandler (ws->old_handler);
	ws->old_handler = NULL;
    }
}

static void
process_error (Ws *ws,
	       XErrorEvent *error)
{
    Trap *trap;
    char buf[64];

    if (!ws->traps)
	g_print ("no traps before popping\n");
    
    pop_obsolete_handlers (ws, error->serial);

    if (!ws->traps)
	g_print ("no traps after popping\n");
    
    for (trap = ws->traps; trap; trap = trap->next)
    {
#if 0
	g_print ("trap: %d %d\n", trap->begin_serial, trap->end_serial);
#endif
	if (trap->begin_serial <= error->serial &&
	    (trap->end_serial > error->serial || trap->end_serial == -1))
	{
	    return;
	}
    }

    XGetErrorText (ws->xdisplay, error->error_code, buf, 63);  
    
    g_error ("Unexpected X error: %s serial %ld error_code "
	     "%d request_code %d minor_code %d)\n",
	     buf,
	     error->serial, 
	     error->error_code, 
	     error->request_code,
	     error->minor_code);
}

static int
xerror_handler (Display		*xdisplay,
		XErrorEvent	*error_event)
{
    GList *list;
    Ws *error_ws;
    
    for (list = all_wses; list != NULL; list = list->next)
    {
	Ws *ws = list->data;

	if (ws->xdisplay == xdisplay)
	{
	    error_ws = ws;
	    break;
	}
    }

    if (error_ws)
    {
	process_error (error_ws, error_event);
    }
    else if (old_x_error_handler)
    {
	(* old_x_error_handler) (xdisplay, error_event);
    }
}

Ws *
ws_new (const char *dpy)
{
    Display *xdisplay = XOpenDisplay (dpy);
    int i;
    Ws *ws;
    int fd;
    
    if (!xdisplay)
	return NULL;
    
    ws = g_new0 (Ws, 1);
    
    ws->xdisplay = xdisplay;

    init_ws (ws);

    fd = ConnectionNumber (xdisplay);
    
    /* process events that are already read off the wire */
    process_events (ws);
    
    fd_add_watch (fd, ws);
    fd_set_read_callback (fd, process_events);
    fd_set_poll_callback (fd, process_events);

    if (!all_wses)
	old_x_error_handler = XSetErrorHandler (xerror_handler);
    
    all_wses = g_list_prepend (all_wses, ws);

    return ws;
}

void
ws_process_xevent (Ws	  *ws,
		   XEvent *xevent)
{
    pop_obsolete_handlers (ws, xevent->xany.serial);
    process_event (ws, xevent);
}

void
ws_begin_error_trap  (Ws	    *ws)
{
    struct Trap *trap = g_new0 (Trap, 1);

    trap->begin_serial = NextRequest (ws->xdisplay);
    trap->end_serial = -1;
    trap->next = ws->traps;
    ws->traps = trap;

    ws->old_handler = XSetErrorHandler (xerror_handler);

#if 0
    g_print ("beginning: %d (%p)\n", trap->begin_serial, ws->xdisplay);
#endif
    g_assert (ws->traps);
}

void
ws_end_error_trap (Ws	    *ws)
{
    Trap *trap;

    trap = ws->traps;
    while (trap && trap->end_serial != -1)
	trap = trap->next;

    if (!trap)
    {
	g_warning ("ws_end_trap_errors() called without corresponding "
		   "ws_begin_trap_errors()");
	return;
    }

    trap->end_serial = NextRequest (ws->xdisplay);

#if 0
    g_print ("ending: %d\n", trap->end_serial);
#endif
    
}

gboolean
ws_init_composite (Ws *ws)
{
    if (XCompositeQueryExtension (
	    ws->xdisplay,
	    &ws->composite.event_base,
	    &ws->composite.error_base))
    {
	ws->composite.available = TRUE;
	
	return TRUE;
    }
    
    return FALSE;
}

gboolean
ws_init_fixes (Ws *ws)
{
    if (XFixesQueryExtension (
	    ws->xdisplay,
	    &ws->composite.event_base,
	    &ws->composite.error_base))
    {
	ws->fixes.available = TRUE;
	
	return TRUE;
    }
    
    return FALSE;
}

gboolean
ws_init_damage (Ws *ws)
{
    if (XDamageQueryExtension (
	    ws->xdisplay,
	    &ws->damage.event_base,
	    &ws->damage.error_base))
    {
	ws->damage.available = TRUE;
	return TRUE;
    }
    
    return FALSE;
}

gboolean
ws_init_shm (Ws *ws)
{
    if (XShmQueryExtension (ws->xdisplay))
    {
	ws->shm.event_base = XShmGetEventBase (ws->xdisplay);
	ws->shm.error_base = -1;
	ws->shm.available = TRUE;
	return TRUE;
    }

    return FALSE;
}

gboolean
ws_init_test (Ws *ws)
{
    int dummy1, dummy2;
    
    if (XTestQueryExtension (
	    ws->xdisplay,
	    &ws->test.event_base,
	    &ws->test.error_base,
	    &dummy1, &dummy2))
    {
	ws->test.available = TRUE;
	return TRUE;
    }

    return FALSE;
}

void
ws_flush (Ws *ws)
{
    g_return_if_fail (ws != NULL);
    
    XFlush (ws->xdisplay);
}

void
ws_sync (Ws *ws)
{
    g_return_if_fail (ws != NULL);
    
    XSync (ws->xdisplay, 0);
}

void
ws_set_synchronize (Ws *ws,
		    gboolean sync)
{
    XSynchronize (ws->xdisplay, sync);
}

WsWindow *
_ws_lookup_window (Ws *ws, XID xid)
{
    g_return_val_if_fail (ws != NULL, NULL);
    
    return g_hash_table_lookup (ws->xresources, (gpointer)xid);
}

gint
ws_screen_get_width (WsScreen *screen)
{
    return screen->xscreen->width;
}

gint
ws_screen_get_height (WsScreen *screen)
{
    return screen->xscreen->height;
}

void
ws_set_ignore_grabs (Ws	      *ws,
		     gboolean  ignore)
{
    g_return_if_fail (ws->test.available);

    XTestGrabControl (ws->xdisplay, ignore);
}

void
_ws_add_resource (Ws *ws, XID xid, WsDrawable *drawable)
{
    g_hash_table_insert (ws->xresources, (gpointer)xid, drawable);
}
